import importlib
from pathlib import Path
import pkgutil

from .entities import (
    Migration,
    MigrationInfo,
    MigrationType,
    NewMigration,
)

from .errors import OperationError


class MigrationsPackageRepo:
    def __init__(self, package: str) -> None:
        self.package = pkgutil.resolve_name(package)
        self.path = Path(*self.package.__path__)

    def get_all(self) -> list[Migration]:
        result = []

        version_packages = [
            p.name
            for p in pkgutil.iter_modules(
                self.package.__path__, f"{self.package.__name__}."
            )
        ]

        for v in version_packages:
            for type_ in MigrationType:
                migrations = self._load_migrations(v, type_)
                result.extend(migrations)

        return result

    @staticmethod
    def _load_migrations(version_path: str, type_: MigrationType) -> list[Migration]:
        version = version_path.split(".")[-1]
        path = f"{version_path}.{type_.name}"

        try:
            package = pkgutil.resolve_name(path)
        except AttributeError:
            return []

        modules = [
            importlib.import_module(m.name)
            for m in pkgutil.iter_modules(package.__path__, f"{path}.")
        ]

        return [
            Migration(
                info=MigrationInfo(
                    version=version,
                    type_=type_,
                    name=m.__name__.split(".")[-1],
                ),
                run=m.run,
                get_progress=getattr(m, "get_progress", None),
            )
            for m in modules
            if hasattr(m, "run")
        ]

    def save(self, migration: NewMigration) -> None:
        version_dir = self.path / migration.info.version

        if not version_dir.exists():
            version_dir.mkdir()
            (version_dir / "__init__.py").touch()

        last_version = sorted(self.path.iterdir())[-1]
        type_dir = last_version / migration.info.type_.name

        if not type_dir.exists():
            type_dir.mkdir()
            (type_dir / "__init__.py").touch()

        migration_path = type_dir / f"{migration.info.name}.py"

        if migration_path.exists():
            raise OperationError("A migration with this name already exists")

        migration_path.write_text(migration.code)



import os

import yaml

from . import cli
from .entities import Config
from .history_repo import HistoryDBRepo
from .migrations_repo import MigrationsPackageRepo
from .migrations_service import MigrationsService
from .operations import ClickHouseConnectOperations


try:
    with open("izbushka.yml") as f:
        conf = yaml.safe_load(f)
except FileNotFoundError:
    cli.init()


config = Config(
    host=os.getenv("CLICKHOUSE_HOST") or conf.get("host") or "localhost",
    port=int(os.getenv("CLICKHOUSE_PORT") or conf.get("port") or 8123),
    username=(os.getenv("CLICKHOUSE_USERNAME") or conf.get("username") or "Default"),
    password=os.getenv("CLICKHOUSE_PASSWORD") or conf.get("password") or "",
    database=os.getenv("CLICKHOUSE_DATABASE") or conf.get("database") or "default",
    interface=os.getenv("CLICKHOUSE_INTERFACE") or conf.get("interface") or "http",
    cluster=os.getenv("CLICKHOUSE_CLUSTER") or conf.get("cluster"),
)

operations = ClickHouseConnectOperations(config)
history_repo = HistoryDBRepo(operations)
migrations_repo = MigrationsPackageRepo(conf["migrations_package"])

migrations_service = MigrationsService(
    operations=operations,
    history_repo=history_repo,
    migrations_repo=migrations_repo,
)

cmd = cli.build_cmd(migrations_service=migrations_service)
cmd()



from pypika import functions as fn  # type: ignore

from . import sql
from .entities import (
    HistoryRecord,
    MigrationInfo,
    MigrationType,
    Status,
)
from .protocols import Operations


class HistoryDBRepo:
    table = sql.Table("izbushka_history")
    local_table = sql.Table("izbushka_history_local")

    def __init__(self, operations: Operations) -> None:
        self.operations = operations

    def initialize(self) -> None:
        cluster = self.operations.config.cluster

        self.operations.command(
            sql.Query.create_table(self.local_table if cluster else self.table)
            .if_not_exists()
            .on_cluster(cluster)
            .columns(
                sql.Column("name", "String"),
                sql.Column("version", "String"),
                sql.Column("type", "String"),
                sql.Column("status", "String"),
                sql.Column("timestamp", "DateTime64(9, 'UTC') DEFAULT now64(9, 'UTC')"),
            )
            .engine("ReplicatedMergeTree" if cluster else "MergeTree")
            .order_by("timestamp")
        )

        if cluster:
            self.operations.command(
                sql.Query.create_table(self.table)
                .if_not_exists()
                .on_cluster(cluster)
                .as_table(
                    f"{self.operations.config.database}."
                    f"{self.local_table.get_table_name()}"
                )
                .engine(
                    "Distributed",
                    cluster,
                    self.operations.config.database,
                    self.local_table.get_table_name(),
                    "rand()",
                )
            )

    def get_all(self) -> list[HistoryRecord]:
        t1 = self.table.as_("t1")
        t2 = self.table.as_("t2")

        max_timestamp_q = (
            sql.Query.from_(t2)
            .select(fn.Max(t2.timestamp).as_("ts"))
            .groupby("name", "version", "type")
        )

        query = (
            sql.Query.from_(t1)
            .select("version", "name", "type", "status")
            .join(max_timestamp_q)
            .on(t1.timestamp == max_timestamp_q.ts)
            .orderby("version", "name", "type")
        )

        result = self.operations.query(query)

        return [
            HistoryRecord(
                info=MigrationInfo(
                    version=version,
                    type_=MigrationType[type_],
                    name=name,
                ),
                status=Status[status],
            )
            for version, name, type_, status in result
        ]

    def save(self, record: HistoryRecord) -> None:
        self.operations.insert(
            self.table,
            [
                (
                    record.info.version,
                    record.info.name,
                    record.info.type_.name,
                    record.status.name,
                )
            ],
            column_names=("version", "name", "type", "status"),
        )



from typing import (
    Any,
    Iterable,
    Protocol,
    Sequence,
    Union,
)

from .entities import (
    Config,
    HistoryRecord,
    Migration,
    NewMigration,
)
from . import sql


class HistoryRepo(Protocol):
    def initialize(self) -> None:
        ...

    def get_all(self) -> list[HistoryRecord]:
        ...

    def save(self, record: HistoryRecord) -> None:
        ...


class MigrationsRepo(Protocol):
    def get_all(self) -> list[Migration]:
        ...

    def save(self, migration: NewMigration) -> None:
        ...


class Operations(Protocol):
    config: Config

    @property
    def client(self) -> Any:
        """Client for interaction with ClickHouse.
        Use this property when methods of the ``Operations``
        class are not enough.
        """
        ...

    def command(self, query: Union[str, sql.Query]) -> Union[str, int, Sequence[str]]:
        """Query the database and get a result as a single value."""
        ...

    def query(self, query: Union[str, sql.Query]) -> Sequence[Sequence]:
        """Query the database and get a result as a sequence of rows."""
        ...

    def insert(
        self,
        table: Union[str, sql.Table],
        data: Sequence[Sequence],
        column_names: Union[str, Iterable[str]] = "*",
    ) -> None:
        """Insert a sequence of rows to a table.

        :param table: Table to insert rows into.
        :param data: Rows to insert.
        :param column_names: Column names."""
        ...



from __future__ import annotations

import enum
from dataclasses import dataclass
from enum import Enum
from typing import (
    Callable,
    NamedTuple,
    Optional,
)


class MigrationType(Enum):
    schema = enum.auto()
    data = enum.auto()
    cleanup = enum.auto()


class MigrationInfo(NamedTuple):
    version: str
    name: str
    type_: MigrationType


class Status(Enum):
    pending = enum.auto()
    in_progress = enum.auto()
    failed = enum.auto()
    done = enum.auto()


@dataclass
class NewMigration:
    default_code = (
        "from izbushka import Operations\n"
        "from izbushka import sql\n"
        "\n\n"
        "def run(op: Operations) -> None:\n"
        "    ...\n"
    )

    info: MigrationInfo
    code: str = default_code


@dataclass
class Migration:
    info: MigrationInfo
    run: Callable
    get_progress: Optional[Callable] = None


@dataclass
class HistoryRecord:
    info: MigrationInfo
    status: Status

    @classmethod
    def in_progress(cls, migration: Migration) -> HistoryRecord:
        return cls(
            info=migration.info,
            status=Status.in_progress,
        )

    @classmethod
    def failed(cls, migration: Migration) -> HistoryRecord:
        return cls(
            info=migration.info,
            status=Status.failed,
        )

    @classmethod
    def done(cls, migration: Migration) -> HistoryRecord:
        return cls(
            info=migration.info,
            status=Status.done,
        )


@dataclass
class Config:
    host: str
    port: int
    username: str
    password: str
    database: str
    interface: str = "http"
    cluster: Optional[str] = None



from typing import Optional

import click
import yaml
from click import Group

from .entities import MigrationType
from .errors import OperationError
from .migrations_service import MigrationsService


@click.command
@click.argument("migrations_package")
def init(migrations_package: str) -> None:
    with click.open_file("izbushka.yml", "w") as f:
        yaml.safe_dump({"migrations_package": migrations_package}, f)


def build_cmd(migrations_service: MigrationsService) -> Group:
    @click.group
    def cmd() -> None:
        pass

    @cmd.command
    def run() -> None:
        click.echo("Applying migrations...", nl=False)
        migrations_service.run()
        click.echo(" [DONE]")

    @cmd.command
    def status() -> None:
        status = migrations_service.get_status()
        click.echo(yaml.safe_dump(status))

    @cmd.command
    @click.argument("name")
    @click.option("--version", "-v")
    @click.option("--type", "-t")
    def add(name: str, version: Optional[str], type: Optional[str]) -> None:
        try:
            migrations_service.add(
                name=name,
                version=version,
                type_=None if type is None else MigrationType[type],
            )
        except OperationError as err:
            click.echo(str(err), err=True)

    return cmd



from typing import Optional

from .entities import (
    HistoryRecord,
    MigrationInfo,
    MigrationType,
    NewMigration,
    Status,
)
from .errors import OperationError
from .protocols import (
    HistoryRepo,
    MigrationsRepo,
    Operations,
)


class MigrationsService:
    def __init__(
        self,
        migrations_repo: MigrationsRepo,
        history_repo: HistoryRepo,
        operations: Operations,
    ) -> None:
        self.migrations_repo = migrations_repo
        self.history_repo = history_repo
        self.operations = operations

    def run(self) -> None:
        self.history_repo.initialize()
        history_records = self.history_repo.get_all()
        done = {r.info for r in history_records}

        migrations = [m for m in self.migrations_repo.get_all() if m.info not in done]

        for migration in migrations:
            self.history_repo.save(HistoryRecord.in_progress(migration))
            try:
                migration.run(self.operations)
            except Exception as ex:
                self.history_repo.save(HistoryRecord.failed(migration))
                raise ex
            else:
                self.history_repo.save(HistoryRecord.done(migration))

    def get_status(self) -> list[dict]:
        self.history_repo.initialize()
        migrations = self.migrations_repo.get_all()
        history_records = self.history_repo.get_all()
        history_dict = {r.info: r for r in history_records}
        result = []

        for m in migrations:
            data = {
                "version": m.info.version,
                "name": m.info.name,
                "type": m.info.type_.name,
            }

            try:
                history_record = history_dict[m.info]
            except KeyError:
                data["status"] = "pending"
            else:
                data["status"] = history_record.status.name
                if (
                    history_record.status == Status.in_progress
                    and m.get_progress is not None
                ):
                    data["progress"] = m.get_progress(self.operations)

            result.append(data)

        return result

    def add(
        self,
        name: str,
        version: Optional[str] = None,
        type_: Optional[MigrationType] = None,
    ) -> None:
        if type_ is None:
            type_ = MigrationType.schema

        if version is None:
            migrations = self.migrations_repo.get_all()

            if not migrations:
                raise OperationError("A version is required for the first migration")

            version = migrations[-1].info.version

        migration = NewMigration(
            info=MigrationInfo(
                name=name,
                version=version,
                type_=type_,
            )
        )

        self.migrations_repo.save(migration)



from typing import (
    Any,
    Iterable,
    Sequence,
    Union,
)

import clickhouse_connect  # type: ignore

from .entities import Config
from . import sql


class ClickHouseConnectOperations:
    def __init__(self, config: Config) -> None:
        self.config = config
        self._client: Any = None

    @property
    def client(self) -> Any:
        if self._client is None:
            self._client = clickhouse_connect.get_client(
                host=self.config.host,
                port=self.config.port,
                username=self.config.username,
                password=self.config.password,
                database=self.config.database,
                interface=self.config.interface,
            )

        return self._client

    def command(self, query: Union[str, sql.Query]) -> Union[str, int, Sequence[str]]:
        q = query if isinstance(query, str) else str(query)
        return self.client.command(q)

    def query(self, query: Union[str, sql.Query]) -> Sequence[Sequence]:
        q = query if isinstance(query, str) else str(query)
        return self.client.query(q).result_set

    def insert(
        self,
        table: Union[str, sql.Table],
        data: Sequence[Sequence],
        column_names: Union[str, Iterable[str]] = "*",
    ) -> None:
        table_name = table if isinstance(table, str) else table.get_table_name()
        self.client.insert(table_name, data, column_names=column_names)



from __future__ import annotations

from typing import (
    Any,
    Optional,
    Sequence,
    Union,
)

from pypika import (  # type: ignore
    Column,
    Table,
    dialects as _dialects,
)

__all__ = ("Column", "Query", "Table")


class Query(_dialects.ClickHouseQuery):
    @classmethod
    def create_table(cls, table: Union[Table, str]) -> ClickHouseCreateQueryBuilder:
        """Entry point for the builder of a CREATE query.
        Takes a name of a table to create."""
        return ClickHouseCreateQueryBuilder().create_table(table)

    @classmethod
    def exchange_tables(cls, table1: str, table2: str) -> Query:
        """Entry point for the builder of an EXCHANGE query.
        Takes names of tables to exchange."""
        return ClickHouseExchangeQueryBuilder().exchange_tables(table1, table2)


class ClickHouseCreateQueryBuilder(_dialects.CreateQueryBuilder):
    QUERY_CLS = Query

    def __init__(self) -> None:
        super().__init__(dialect=_dialects.Dialects.CLICKHOUSE)
        self._cluster: Optional[str] = None
        self._engine: Optional[str] = None
        self._engine_params: Sequence[str] = ()
        self._order_by: Optional[str] = None
        self._as_table: Optional[str] = None

    @_dialects.builder
    def on_cluster(self, cluster: str) -> ClickHouseCreateQueryBuilder:  # type: ignore
        if self._cluster:
            raise AttributeError("'CreateQuery' object already has attribute cluster")
        self._cluster = cluster

    @_dialects.builder
    def engine(  # type: ignore
        self, engine: str, *args: str
    ) -> ClickHouseCreateQueryBuilder:
        if self._engine:
            raise AttributeError("'CreateQuery' object already has attribute engine")
        self._engine = engine
        self._engine_params = args

    @_dialects.builder
    def order_by(self, order: str) -> ClickHouseCreateQueryBuilder:  # type: ignore
        if self._order_by:
            raise AttributeError("'CreateQuery' object already has attribute order_by")
        self._order_by = order

    @_dialects.builder
    def as_table(self, table: str) -> ClickHouseCreateQueryBuilder:  # type: ignore
        if self._as_table:
            raise AttributeError("'CreateQuery' object already has attribute as_table")
        self._as_table = table

    def get_sql(self, **kwargs: Any) -> str:
        self._set_kwargs_defaults(kwargs)

        if not self._create_table:
            return ""

        if not any((self._columns, self._as_select, self._as_table)):
            return ""

        create_table = self._create_table_sql(**kwargs)

        if self._as_table:
            query = f"{create_table} AS {self._as_table}"
        elif self._as_select:
            query = create_table + self._as_select_sql(**kwargs)
        else:
            body = self._body_sql(**kwargs)
            table_options = self._table_options_sql(**kwargs)

            query = "{create_table} ({body}){table_options}".format(
                create_table=create_table, body=body, table_options=table_options
            )

        if self._engine is not None:
            query = f"{query} ENGINE = {self._engine}"
            if self._engine_params:
                query = f"{query}({', '.join(self._engine_params)})"

        if self._order_by is not None:
            query = f"{query} ORDER BY {self._order_by}"

        return query

    def _create_table_sql(self, **kwargs: Any) -> str:
        query = super()._create_table_sql(**kwargs)
        if self._cluster:
            query = f"{query} ON CLUSTER {self._cluster}"
        return query


class ClickHouseExchangeQueryBuilder:
    def __init__(self) -> None:
        self._table1: Optional[str] = None
        self._table2: Optional[str] = None
        self._cluster: Optional[str] = None

    @_dialects.builder
    def exchange_tables(  # type: ignore
        self,
        table1: str,
        table2: str,
    ) -> ClickHouseExchangeQueryBuilder:
        self._table1 = table1
        self._table2 = table2

    @_dialects.builder
    def on_cluster(  # type: ignore
        self, cluster: str
    ) -> ClickHouseExchangeQueryBuilder:
        if self._cluster:
            raise AttributeError("'ExchangeQuery' object already has attribute cluster")

        self._cluster = cluster

    def get_sql(self, **kwargs: Any) -> str:
        query = f"EXCHANGE TABLES {self._table1} AND {self._table2}"

        if self._cluster is not None:
            query += f" ON CLUSTER {self._cluster}"

        return query

    def __str__(self) -> str:
        return self.get_sql()

    def __repr__(self) -> str:
        return self.__str__()
